;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "julia-vterm" "20260813.2350"
  "A mode for Julia REPL using vterm."
  '((emacs "25.1")
    (vterm "0.0.1"))
  :url "https://github.com/shg/julia-vterm.el"
  :commit "718b3952e980f73eb33b67168145c383b1a5f0bf"
  :revdesc "718b3952e980"
  :keywords '("languages" "julia"))
